# -------------------------------------------------------------------------------------------------
# test_lock1_mp_xi2.py: Multiprocessing, Pipes and Queues with Lock                              
#
# Theory:      Synchronization between processes
#
# Description: Multiprocessing contains equivalents of all the synchronization primitives from threading. 
#              For instance one can use a lock to ensure that only one process prints to standard output at a time
#
#              The Pipe() function returns a pair of connection objects connected by a pipe which by default is duplex (two-way). 
#              The two connection objects returned by Pipe() represent the two ends of the pipe. 
#              Each connection object has send() and recv() methods (among others). 
#              Note that data in a pipe may become corrupted if two processes (or threads) try to read from or write to the same end of the pipe at the same time. 
#              Of course there is no risk of corruption from processes using different ends of the pipe at the same time.
#
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Imports from standard modules                              
# -------------------------------------------------------------------------------------------------
import os
import sys


# -------------------------------------------------------------------------------------------------
# Imports from multiprocessing package                              
# -------------------------------------------------------------------------------------------------
# from multiprocessing import Process, Queue, Pipe, Lock
from multiprocessing.process import Process
from multiprocessing.queues  import Queue
from multiprocessing.connection import Pipe
from multiprocessing.synchronize import Lock


# -------------------------------------------------------------------------------------------------
# yourDragonXi imports                               
# -------------------------------------------------------------------------------------------------
from process2xi import Process2Xi
from queue2xi  import Queue2Xi
from pipe2xi import Pipe2Xi
from lock2xi import Lock2Xi

# -------------------------------------------------------------------------------------------------
# sense function for multiprocessing
# -------------------------------------------------------------------------------------------------
def sense(target_param, queue_param, pipe_param,lock_param):
    target_xi2 = target_param
    queue_xi2 = queue_param
    pipe_publisher_channel_xi2 = pipe_param
    lock_xi2 = lock_param
    
    print("SENSE: ", target_xi2)
    # ---------------------------------------------------------------------------------------------
    # put a message to the queue
    # 
    queue_xi2.put(['DragonXi2_00', target_xi2, 'sensing!']) # puts to its queue
    # ---------------------------------------------------------------------------------------------
    # 
    # ---------------------------------------------------------------------------------------------
    #  
    lock_xi2.acquire()
    # 
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_01', 'BIO ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_02', 'CO2 ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_03', 'Coal ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_04', 'Diesel ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_05', 'Electricity ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_06', 'Fusion ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_07', 'Gas ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_08', 'Geo ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_09', 'Hydrogen ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_10', 'Lithium ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_11', 'Nuclear Power ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_12', 'Ocean ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_13', 'Oil ',target_xi2]) 
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_14', 'Solar ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_15', 'Wind ',target_xi2])
    #
    lock_xi2.release()
    # ---------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# reason funcion for multiprocessing
# -------------------------------------------------------------------------------------------------
def reason(target_param, queue_param, pipe_param,lock_param):
    target_xi4 = target_param
    queue_xi4 = queue_param
    pipe_publisher_channel_xi4 = pipe_param
    lock_xi4 = lock_param
    
    print("REASON: ", target_xi4)
    # ---------------------------------------------------------------------------------------------
    # put a message to the queue
    # 
    queue_xi4.put(['DragonXi4_00', target_xi4, 'portfolio must be sustainable!']) # puts to its queue 
    # ---------------------------------------------------------------------------------------------
    # reserve the pipe using the lock, send messages and release the pipe
    # 
    lock_xi4.acquire()
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_01', 'BIO ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_02', 'Geo ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_03', 'Hydrogen ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_04', 'Nuclear Power ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_05', 'Ocean ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_06', 'Solar ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_07', 'Wind ',target_xi4])    
    lock_xi4.release()
        
# -------------------------------------------------------------------------------------------------
# ace function for multiprocessing
# -------------------------------------------------------------------------------------------------
def ace(target_param, queue_param, pipe_param,lock_param):
    # ---------------------------------------------------------------------------------------------
    # 
    # 
    target_xi8 = target_param
    queue_xi8 = queue_param
    pipe_publisher_channel_xi8 = pipe_param
    lock_xi8 = lock_param
    
    print("ACE: ", target_xi8)
    # ---------------------------------------------------------------------------------------------
    # put a message to the queue
    # 
    queue_xi8.put(['DragonXi8_00', target_xi8, 'to be aced!']) # puts to its queue       
    
    # ---------------------------------------------------------------------------------------------
    # reserve the pipe using the lock, send messages and release the pipe
    # 
    lock_xi8.acquire()
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_01', 'BIO ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_02', 'Hydrogen ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_03', 'Nuclear Power ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_04', 'Solar ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_05', 'Wind ',target_xi8])    
    lock_xi8.release()

# -------------------------------------------------------------------------------------------------
# phase1: testing with standard multiprocessing                              
# -------------------------------------------------------------------------------------------------    
def phase1():
    # =============================================================================================
    # PHASE I: use the original multiprocessing modules
    # =============================================================================================
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing queues for sensing, reasoning and acing
    # 
    queue2xi = Queue()
    queue4xi = Queue()
    queue8xi = Queue()
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing pipes with consumer and publisher ends
    #     
    pipe2xi_consumer_channel, pipe2xi_publisher_channel = Pipe()
    pipe4xi_consumer_channel, pipe4xi_publisher_channel = Pipe()
    pipe8xi_consumer_channel, pipe8xi_publisher_channel = Pipe()
    # ---------------------------------------------------------------------------------------------
    # create three locks
    #   
    lock_xi2 = Lock()
    lock_xi4 = Lock()
    lock_xi8 = Lock()    
    
    # ---------------------------------------------------------------------------------------------
    # SENSE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    #
    print("\n~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~")
    process2xi = Process(target=sense, args=('Energy ',queue2xi, pipe2xi_publisher_channel,lock_xi2))
    process2xi.start()
    print (queue2xi.get())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    process2xi.join()
    # ---------------------------------------------------------------------------------------------
    # REASON IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~")
    process4xi = Process(target=reason, args=('Energy ',queue4xi, pipe4xi_publisher_channel,lock_xi4))
    process4xi.start()
    print (queue4xi.get())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    process4xi.join()
    # ---------------------------------------------------------------------------------------------
    # ACE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~")
    process8xi = Process(target=ace, args=('Energy ',queue8xi, pipe8xi_publisher_channel,lock_xi8))
    process8xi.start()
    print (queue8xi.get())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    process8xi.join()
    # ---------------------------------------------------------------------------------------------
    # Notes:
    #       above assumed the main to know amount of messages
    #       no locks were used
    print("\n===============================================================")
    
# -------------------------------------------------------------------------------------------------
# phase2: testing with dragonized multiprocessing                              
# -------------------------------------------------------------------------------------------------    
def phase2():    
    # =============================================================================================
    # PHASE II: using yourDragonXi modules
    # =============================================================================================
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing queues for sensing, reasoning and acing
    # 
    queue2xi = Queue2Xi()
    queue4xi = Queue2Xi()
    queue8xi = Queue2Xi()
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing pipes with consumer and publisher ends
    #     
    pipe2xi_consumer_channel, pipe2xi_publisher_channel = Pipe2Xi()
    pipe4xi_consumer_channel, pipe4xi_publisher_channel = Pipe2Xi()
    pipe8xi_consumer_channel, pipe8xi_publisher_channel = Pipe2Xi()
    # ---------------------------------------------------------------------------------------------
    # create three locks
    #   
    lock_xi2 = Lock2Xi()
    lock_xi4 = Lock2Xi()
    lock_xi8 = Lock2Xi()    
    
    # ---------------------------------------------------------------------------------------------
    # SENSE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    #
    print("\n with yourDragonXi!")
    process2xi = Process2Xi(target=sense, args=('Energy ',queue2xi, pipe2xi_publisher_channel,lock_xi2))
    process2xi.start()
    print (queue2xi.get())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    process2xi.join()
    # ---------------------------------------------------------------------------------------------
    # REASON IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n with yourDragonXi!")
    process4xi = Process2Xi(target=reason, args=('Energy ',queue4xi, pipe4xi_publisher_channel,lock_xi4))
    process4xi.start()
    print (queue4xi.get())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    process4xi.join()
    # ---------------------------------------------------------------------------------------------
    # ACE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n with yourDragonXi!")
    process8xi = Process2Xi(target=ace, args=('Energy ',queue8xi, pipe8xi_publisher_channel,lock_xi8))
    process8xi.start()
    print (queue8xi.get())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    process8xi.join()
    # ---------------------------------------------------------------------------------------------
    # Notes:
    #       above assumed the main to know amount of messages
    #       no locks were used
    print("\n===============================================================")  

    
# -------------------------------------------------------------------------------------------------
# main                              
# -------------------------------------------------------------------------------------------------    
def main():
    phase1()
    phase2()
            
# -------------------------------------------------------------------------------------------------
#                               
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    main()

# -------------------------------------------------------------------------------------------------
# Output of test_lock1_mp_xi2.py                             
# -------------------------------------------------------------------------------------------------
# ~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~
# ['DragonXi2_00', 'Energy ', 'sensing!']
# ['DragonXi2_Sensing_01', 'BIO ', 'Energy ']
# ['DragonXi2_Sensing_02', 'CO2 ', 'Energy ']
# ['DragonXi2_Sensing_03', 'Coal ', 'Energy ']
# ['DragonXi2_Sensing_04', 'Diesel ', 'Energy ']
# ['DragonXi2_Sensing_05', 'Electricity ', 'Energy ']
# ['DragonXi2_Sensing_06', 'Fusion ', 'Energy ']
# ['DragonXi2_Sensing_07', 'Gas ', 'Energy ']
# ['DragonXi2_Sensing_08', 'Geo ', 'Energy ']
# ['DragonXi2_Sensing_09', 'Hydrogen ', 'Energy ']
# ['DragonXi2_Sensing_10', 'Lithium ', 'Energy ']
# ['DragonXi2_Sensing_11', 'Nuclear Power ', 'Energy ']
# ['DragonXi2_Sensing_12', 'Ocean ', 'Energy ']
# ['DragonXi2_Sensing_13', 'Oil ', 'Energy ']
# ['DragonXi2_Sensing_14', 'Solar ', 'Energy ']
# ['DragonXi2_Sensing_15', 'Wind ', 'Energy ']
# 
# ~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~
# ['DragonXi4_00', 'Energy ', 'portfolio must be sustainable!']
# ['DragonXi4_Reasoning_01', 'BIO ', 'Energy ']
# ['DragonXi4_Reasoning_02', 'Geo ', 'Energy ']
# ['DragonXi4_Reasoning_03', 'Hydrogen ', 'Energy ']
# ['DragonXi4_Reasoning_04', 'Nuclear Power ', 'Energy ']
# ['DragonXi4_Reasoning_05', 'Ocean ', 'Energy ']
# ['DragonXi4_Reasoning_06', 'Solar ', 'Energy ']
# ['DragonXi4_Reasoning_07', 'Wind ', 'Energy ']
# 
# ~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~
# ['DragonXi8_00', 'Energy ', 'to be aced!']
# ['DragonXi8_Ace_01', 'BIO ', 'Energy ']
# ['DragonXi8_Ace_02', 'Hydrogen ', 'Energy ']
# ['DragonXi8_Ace_03', 'Nuclear Power ', 'Energy ']
# ['DragonXi8_Ace_04', 'Solar ', 'Energy ']
# ['DragonXi8_Ace_05', 'Wind ', 'Energy ']
# 
# ===============================================================
# 
#  with yourDragonXi!
# ['DragonXi2_00', 'Energy ', 'sensing!']
# ['DragonXi2_Sensing_01', 'BIO ', 'Energy ']
# ['DragonXi2_Sensing_02', 'CO2 ', 'Energy ']
# ['DragonXi2_Sensing_03', 'Coal ', 'Energy ']
# ['DragonXi2_Sensing_04', 'Diesel ', 'Energy ']
# ['DragonXi2_Sensing_05', 'Electricity ', 'Energy ']
# ['DragonXi2_Sensing_06', 'Fusion ', 'Energy ']
# ['DragonXi2_Sensing_07', 'Gas ', 'Energy ']
# ['DragonXi2_Sensing_08', 'Geo ', 'Energy ']
# ['DragonXi2_Sensing_09', 'Hydrogen ', 'Energy ']
# ['DragonXi2_Sensing_10', 'Lithium ', 'Energy ']
# ['DragonXi2_Sensing_11', 'Nuclear Power ', 'Energy ']
# ['DragonXi2_Sensing_12', 'Ocean ', 'Energy ']
# ['DragonXi2_Sensing_13', 'Oil ', 'Energy ']
# ['DragonXi2_Sensing_14', 'Solar ', 'Energy ']
# ['DragonXi2_Sensing_15', 'Wind ', 'Energy ']
# 
#  with yourDragonXi!
# ['DragonXi4_00', 'Energy ', 'portfolio must be sustainable!']
# ['DragonXi4_Reasoning_01', 'BIO ', 'Energy ']
# ['DragonXi4_Reasoning_02', 'Geo ', 'Energy ']
# ['DragonXi4_Reasoning_03', 'Hydrogen ', 'Energy ']
# ['DragonXi4_Reasoning_04', 'Nuclear Power ', 'Energy ']
# ['DragonXi4_Reasoning_05', 'Ocean ', 'Energy ']
# ['DragonXi4_Reasoning_06', 'Solar ', 'Energy ']
# ['DragonXi4_Reasoning_07', 'Wind ', 'Energy ']
# 
#  with yourDragonXi!
# ['DragonXi8_00', 'Energy ', 'to be aced!']
# ['DragonXi8_Ace_01', 'BIO ', 'Energy ']
# ['DragonXi8_Ace_02', 'Hydrogen ', 'Energy ']
# ['DragonXi8_Ace_03', 'Nuclear Power ', 'Energy ']
# ['DragonXi8_Ace_04', 'Solar ', 'Energy ']
# ['DragonXi8_Ace_05', 'Wind ', 'Energy ']
# 
# ===============================================================
#   
# -------------------------------------------------------------------------------------------------
# end of test_lock1_mp_xi2.py                              
# -------------------------------------------------------------------------------------------------
 